#include "sortiva.hpp"

#include <raylib.h>

#include <cmath>


constexpr Color sorter_colors[] = {
	{ 253, 249, 0, 255 },	// Yellow
	{ 255, 203, 0, 255 },	// Gold
	{ 255, 161, 0, 255 },	// Orange
	{ 255, 109, 194, 255 },	// Pink
	{ 230, 41, 55, 255 },	// Red
	{ 190, 33, 55, 255 },	// Maroon
	{ 0, 228, 48, 255 },		// Green
	{ 0, 158, 47, 255 },		// Lime
	{ 0, 117, 44, 255 },		// Dark Green
	{ 102, 191, 255, 255 },	// Sky Blue
	{ 0, 121, 241, 255 },	// Blue
	{ 0, 82, 172, 255 },		// Dark Blue
	{ 255, 0, 255, 255 },	// Magenta
	{ 200, 122, 255, 255 },	// Purple
	{ 135, 60, 190, 255 },	// Violet
	{ 112, 31, 126, 255 },	// Dark Purple
	{ 211, 176, 131, 255 },	// Beige
	{ 127, 106, 79, 255 },	// Brown
	{ 76, 63, 47, 255 },		// Dark Brown
};

constexpr int sorter_colors_size = sizeof(sorter_colors) / sizeof(Color);

constexpr Color sorter_block_colors[] = {
	{ 200, 200, 200, 255 },	// Light Gray
	{ 80, 80, 80, 255 },		// Dark Gray
};

constexpr int sorter_block_colors_size = sizeof(sorter_block_colors) / sizeof(Color);


void Sortiva::draw()
{
	Vector2 margin = {
		.x = m_Width * 0.01f,
		.y = 70
	};
	DrawRectangleV(margin, { m_Width - 2 * margin.x, m_Height - 2 * margin.y }, { 60, 60, 60, 255 });


	margin = {
		.x = m_Width * 0.02f,
		.y = 80
	};

	Rectangle plane = {
		.x = margin.x,
		.y = margin.y,
		.width = m_Width - margin.x * 2,
		.height = m_Height - margin.y * 2
	};

	if (m_Steps->empty())
	{
		static const char* msg = "Press \"Run\" to start a sorting";
		static int width = MeasureText(msg, 20);
		DrawText(msg, plane.x + plane.width / 2 - width / 2, plane.y + plane.height / 2 - 10, 20, { 245, 245, 245, 255 });
		return;
	}

	// +2 for start points and end points; will be removed later in the function
	size_t cvals = m_Steps->size();

	size_t steps = m_Steps->at(0).depth();
	for (size_t i = 1; i < cvals; ++i)
	{
		steps = std::min(steps, m_Steps->at(i).depth());
	}

	steps += 1;

	// ui distance values
	int w = static_cast<int>(plane.width / static_cast<float>(steps));
	int g = plane.height / static_cast<float>(cvals + 1);
	int h = (plane.height - static_cast<float>(cvals - 1) * g) / 2;
	int e = w / 2;
	float wf = static_cast<float>(w);

	float pwhd = sqrtf(plane.width * plane.width + plane.height * plane.height);
	float pw = plane.width / pwhd - 0.01f;
	float ph = plane.height / pwhd;
	float l = pw * ph * pwhd * 0.02f - static_cast<float>(steps) / 2; // Make more fitting

	for (size_t i = 0; i < steps; ++i)
	{
		DrawRectangle(
			static_cast<int>(i) * w + static_cast<int>(plane.x),
			plane.y,
			w,
			plane.height,
			sorter_block_colors[i % sorter_block_colors_size]
		);
	}
	steps -= 1;

	//  colors to values - ratio ; adds high variance to the color selection
	int colid = sorter_colors_size / cvals;

	// font settings
	constexpr Color textColor = { 0, 0, 0, 255 };
	int l3 = static_cast<int>(l * 3);

#ifdef _DEBUG
	DrawFPS(5, 5);
#endif

	for (int v = 0; v < cvals; ++v)
	{
		linked_list<uint16_t>* list = &m_Steps->at(v);
		uint16_t value = list->value;
		Color col = sorter_colors[(v * colid) % sorter_colors_size];




		int s = 0;
		for (; list->next; list = list->next)
		{
			float sf = static_cast<float>(s);

			value = list->value;
			uint16_t nvalue = list->next->value;
			float y = plane.y + h + g * (value - 1);
			float yt = plane.y + h + g * (nvalue - 1);


			DrawSplineSegmentBezierCubic(
				{ (sf + 1.f) * wf + plane.x - 1, y },
				{ (sf + 2.f) * wf + plane.x, y },
				{ (sf + 1.f) * wf + plane.x, yt },
				{ (sf + 2.f) * wf + plane.x, yt },
				l,
				col
			);

#ifdef _DEBUG
			float font_size = l * 1.2f;
			int yd = y - yt < 0 ? -font_size : +font_size;
			yd *= 2;
			DrawText(
				TextFormat("%d -> %d", value, nvalue),
				(sf + 1.f) * wf + plane.x + 10,
				y + yd,
				font_size,
				col);
#endif
			++s;
		}

		value = list->value;

		Vector2 pos = { wf * steps + e + plane.x, h + (value - 1) * g + plane.y };
		DrawCircleV(pos, l3, col);

		DrawLineEx(
			{ pos.x - e - 1, pos.y },
			pos,
			l,
			col
		);

		const char* strv = TextFormat("%d", v + 1);
		DrawText(
			strv,
			static_cast<int>(pos.x) + e / 2,
			static_cast<int>(pos.y) - l3 / 2,
			l3,
			textColor
		);

		if (list->first)
			list = list->first;
		value = list->value;
		pos = { wf - e + plane.x, h + (value - 1) * g + plane.y };

		DrawText(
			strv,
			static_cast<int>(pos.x) - e / 2 - l3 / 2,
			static_cast<int>(pos.y) - l3 / 2,
			l3,
			textColor
		);

		DrawRing(
			pos,
			l * 2,
			l3,
			20.f,
			340.f,
			10,
			col
		);

		DrawLineEx(
			pos,
			{ pos.x + e, pos.y },
			l,
			col
		);
	}

}